//
//  MiscSplitView.h
//
//  Written by Dwight D. Everhart.  Based on SideSplitView by Mary McNabb
//  of NeXT (from the ZooView mini-example).
//
//  Copyright (c) 1996, 1997 by Dwight D. Everhart.  All rights reserved.
//  Version 1.42; April 4, 1997
//
//      This notice may not be removed from this source code.
//
//  This object is included in the MiscKit by permission from the author
//  and its use is governed by the MiscKit license, found in the file
//  "LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//  for a list of all applicable permissions and restrictions.
//
//  Although this code started as SideSplitView, there is little that remains
//  of the original code.  Still, the author is grateful to Ms. McNabb for
//  providing a starting point for this code.  This is the distribution notice
//  that she put in her code:
//
//    You may freely copy, distribute, and reuse the code in this example.
//    NeXT disclaims any warranty of any kind, expressed or implied, as to its
//    fitness for any particular use.
//
//  Purpose:
//    This class implements a view which divides its frame, either vertically
//    or horizontally, into two or more adjacent areas in which are displayed
//    its subviews.  The user can change the proportion of the areas occupied
//    by the subviews relative to each other.  The class can also save and
//    restore its subview divisions to a string or a default.  The interface
//    is compatible with NXSplitView, and the divisions saving mechanism has
//    the same style of interface as Window's frame saving mechanism.
//
//  Requires:
//    MiscAppDefaults Application category
//    Dimple.tiff image (embedded into Dimple_tiff.m by Don Yacktman)
//
//  Release notes:
//    Version 1.11:  February 5, 1996
//      First release
//    Version 1.20:  April 18, 1996
//      Removed check for autosave name uniqueness.  This check caused
//      problems when an application used multiple MiscSplitViews with
//      the same autosave name, which would be the case if a MiscSplitView
//      were used as part of a document window in a multiple-document
//      architecture:  only the MiscSplitView in the first window opened
//      would have its divider positions saved in the defaults database;
//      the others would not.  If you prefer to have the check enabled,
//      you can enable it with the requireUniqueAutosaveNames: class method.
//    Version 1.21:  April 22, 1996
//      Fixed two bugs related to the autosave name hash table:
//      1. The removeDividersUsingName: class method incorrectly tried to
//         remove the hash table entries for the specified name by using
//         that name as the key.  Names are actually the values in the table;
//         the keys are the MiscSplitView objects.  The offending line was
//         removed.  Now no attempt is made to remove hash table entries
//         related to the name; only the default is removed from the user's
//         defaults database.
//      2. When a MiscSplitView was freed, its entry in the hash table was
//         not removed.  Now it is, in the newly added free method.  This
//         bug caused crashes when the hash table sent the freed objects
//         messages during rehashing, which occurs when the table is enlarged.
//    Version 1.30:  October 3, 1996
//      Fixed a bug that caused the subviews to be incorrectly sized if there
//      were more than two of them.  Thanks to Champciaux Laurent for finding
//      this bug.
//    Version 1.40:  February 27, 1997
//      Streamlined the model event loop in mouseDown:.  It also now
//      explicitly enables event coalescing, and it now uses a timeout so
//      that if something weird happens and the loop doesn't notice that the
//      dragging has stopped, the timeout will cause the loop to exit.
//    Version 1.41:  March 31, 1997
//      Fixed a bug which caused the left (or top) subview to be resized to a
//      width (or height) of zero if the divider was clicked on without being
//      moved.
//    Version 1.42:  April 4, 1997
//      Increased the defaultName string sizes from 64 to 128.
//

#import <appkit/appkit.h>

@interface MiscSplitView: View

{
   id delegate;
   struct _msvFlags {
#ifdef __BIG_ENDIAN__
      unsigned int horizontal:1;
      unsigned int needsRetiling:1;
      unsigned int _RESERVED:6;
#else
      unsigned int _RESERVED:6;
      unsigned int needsRetiling:1;
      unsigned int horizontal:1;
#endif
   } msvFlags;
}

/* Initializing a MiscSplitView */
+ initialize;
- initFrame: (const NXRect *) frameRect;

/* Freeing a MiscSplitView */
- free;

/* Assigning a delegate */
- delegate;
- setDelegate: anObject;

/* Assigning orientation */
- (BOOL) isHorizontal;
- setHorizontal: (BOOL) makeHorizontal;

/* Accessing the dividers */
- setDividersFromString:   (const char *) data;
- saveDividersToString:    (char *) string;
- setDividersUsingName:    (const char *) name;
- saveDividersUsingName:   (const char *) name;
+ removeDividersUsingName: (const char *) name;
- (const char *) dividersAutosaveName;
- setDividersAutosaveName: (const char *) name;
+ requireUniqueAutosaveNames: (BOOL) require;

/* Drawing the view */
- (NXCoord) dividerSize;
- (NXCoord) dividerHeight;
- (NXCoord) dividerWidth;
- drawSelf: (const NXRect *) rects : (int) rectCount;
- drawDivider: (const NXRect *) aRect;

/* Managing component views */
- constrainDivider: (int) dividerIndex
  leftFrame:  (const NXRect *) leftFrame
  rightFrame: (const NXRect *) rightFrame
  min: (NXCoord *) min max: (NXCoord *) max;
- constrainFramesLeft: (NXRect *) leftFrame andRight: (NXRect *) rightFrame
  besideDivider: (int) dividerIndex;
- adjustSubviews;
- resizeSubviews: (const NXSize *) oldSize;
- setAutoresizeSubviews: (BOOL) flag;
- addSubview: aView;
- addSubview: aView : (int) place relativeTo: otherView;
- replaceSubview: oldView with: newView;

/* Handling events */
- (BOOL) acceptsFirstMouse;
- mouseDown: (NXEvent *) theEvent;

/* Archiving */
- write: (NXTypedStream *) stream;
- read:  (NXTypedStream *) stream;

@end


@interface Object (MiscSplitViewDelegate)

- splitView: sender resizeSubviews: (const NXSize *) oldSize;
- splitView: sender getMinY: (NXCoord *) minY maxY: (NXCoord *) maxY
  ofSubviewAt: (int) offset;
- splitView: sender getMinX: (NXCoord *) minX maxX: (NXCoord *) maxX
  ofSubviewAt: (int) offset;
- splitViewDidResizeSubviews: sender;

@end

