//=============================================================================
//
//		Copyright (C) 1995-1997 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//		
//=============================================================================
//-----------------------------------------------------------------------------
// MiscDrawList.cc
//
//		An extensible array of specifications for drawing cell contents.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscDrawList.cc,v 1.4 97/02/11 19:16:43 sunshine Exp $
// $Log:		MiscDrawList.cc,v $
// Revision 1.4	 97/02/11  19:16:43	 sunshine
// v123: Fixed bug: Was unconditionally using screenFont even when not drawing
// to screen (such as during printing, copying).
// 
// Revision 1.3	 97/02/04  08:24:37	 sunshine
// v121: Fixed #import/#include (<...> vs. "...") so compiles for MiscKit.
// 
// Revision 1.2	 96/09/13  12:35:04	 zarnuk
// Fixed bug.  Was not using screenFont to perform calculations
// when it was using screenFont to perform drawing.
//-----------------------------------------------------------------------------
#ifdef __GNUC__
#pragma implementation
#endif
#include "MiscDrawList.h"
#include "MiscLineWrapper.h"
#include <misckit/MiscTableCell.h>

extern "Objective-C" {
#import <appkit/Cell.h>
#import <appkit/Font.h>
#import <appkit/NXImage.h>
#import <appkit/Text.h>			// NX_LEFTALIGNED
}

extern "C" {
#include <math.h>
#include <stdlib.h>
}

//-----------------------------------------------------------------------------
// get_char_wrap
//-----------------------------------------------------------------------------
inline static bool get_char_wrap( id cell )
	{
	if ([cell isKindOf:[Cell class]])
		{
		struct CellDefs { @defs(Cell) };
		return ((CellDefs const*) cell)->cFlags2.noWrap;
		}
	return false;
	}


//-----------------------------------------------------------------------------
// append
//-----------------------------------------------------------------------------
void MiscDrawList::append( NXRect frame_rect, id cell, BOOL lit,
						NXColor default_text_color,
						Font* default_font )
	{
	char const* txt = 0;
	NXImage* img = 0;

	if (cell != 0)
		{
		if ([cell respondsTo:@selector(image)])
			img = [cell image];
		else if ([cell respondsTo:@selector(icon)])
			{
			char const* nm = [cell icon];
			if (nm != 0)
				img = [NXImage findImageNamed:nm];
			}

		if (img == 0)
			if ([cell respondsTo:@selector(title)])
				txt = [cell title];
			else if ([cell respondsTo:@selector(stringValue)])
				txt = [cell stringValue];

		if (img != 0 || (txt != 0 && *txt != 0))
			{
			if (num_recs >= max_recs)
				{
				max_recs += max_recs;
				recs = (Rec*) realloc( recs, max_recs * sizeof(*recs) );
				}
			Rec& r = recs[ num_recs++ ];

			if (img != 0)
				{
				r.image = img;
				r.icon_rect = frame_rect;
				if ([cell respondsTo:@selector(getIconRect:)])
					[cell getIconRect:&(r.icon_rect)];
				}
			else
				r.image = 0;

			if (txt != 0 && *txt != 0)
				{
				int const len = strlen( txt ) + 1;
				r.text = (char*) malloc( len );
				memcpy( r.text, txt, len );

				r.text_rect = frame_rect;
				if ([cell respondsTo:@selector(getTitleRect:)])
					[cell getTitleRect:&(r.text_rect)];

				r.font = 0;
				if ([cell respondsTo:@selector(font)])
					r.font = [cell font];
				if (r.font == 0)
					r.font = default_font;

				NXColor color;
				if (lit && [cell respondsTo:@selector(highlightTextColor)])
					color = [cell highlightTextColor];
				else if ([cell respondsTo:@selector(textColor)])
					color = [cell textColor];
				else
					color = default_text_color;

				r.text_color = color_list.store( color );

				int flags;
				if ([cell respondsTo:@selector(alignment)])
					flags = ([cell alignment] << 1);
				else
					flags = (NX_LEFTALIGNED << 1);
				if (get_char_wrap( cell ))
					flags |= 1;
				r.flags = flags;
				}
			else
				{
				r.text = 0;
				r.font = 0;
				r.text_color = -1;
				}
			}
		}
	}


//-----------------------------------------------------------------------------
// Rec::draw_icon
//-----------------------------------------------------------------------------
void MiscDrawList::Rec::draw_icon()
	{
	NXPoint p;
	p.x = icon_rect.origin.x;
	p.y = icon_rect.origin.y + icon_rect.size.height;

	NXRect r;
	r.origin.x = 0;
	r.origin.y = 0;
	[image getSize:&(r.size)];

	float const dx = icon_rect.size.width - r.size.width;
	if (dx < 0)
		r.size.width = icon_rect.size.width;
	else if (dx > 0)
		p.x = floor( p.x + dx / 2 );

	float const dy = icon_rect.size.height - r.size.height;
	if (dy < 0)
		r.size.height = icon_rect.size.height;
	else if (dy > 0)
		p.y = floor( p.y - dy / 2 );

	[image composite:NX_SOVER fromRect:&r toPoint:&p];
	}


//-----------------------------------------------------------------------------
// Rec::draw_text
//-----------------------------------------------------------------------------
void MiscDrawList::Rec::draw_text( MiscLineWrapper& lw ) const
	{
	lw.setText( text );
	lw.setRect( text_rect );
	lw.setCharWrap( flags & 1 );
	lw.setAlignment( flags >> 1 );
	lw.draw();
	}


//-----------------------------------------------------------------------------
// draw_icons
//-----------------------------------------------------------------------------
void MiscDrawList::draw_icons()
	{
	for (int i = 0; i < num_recs; i++)
		{
		Rec& r = recs[i];
		if (r.image != 0)
			r.draw_icon();
		}
	}


//-----------------------------------------------------------------------------
// draw_text
//-----------------------------------------------------------------------------
void MiscDrawList::draw_text()
	{
	MiscLineWrapper lw;
	lw.setNoPartialLines( true );
	for (int i = 0; i < num_recs; i++)			// For each font...
		{
		if (recs[i].font != 0)
			{
			Font* const font = recs[i].font;
			Font* screenFont;
			if (NXDrawingStatus == NX_DRAWING &&
				(screenFont = [font screenFont]) != 0)
				{
				[screenFont set];
				lw.setFont( screenFont );
				}
			else
				{
				[font set];
				lw.setFont( font );
				}

			for (int j = i; j < num_recs; j++)	// For each font-color pair...
				{
				if (recs[j].text_color >= 0)
					{
					int const color_id = recs[j].text_color;
					NXSetColor( color_list[ color_id ] );

					for (int k = j; k < num_recs; k++)	// For each record...
						{
						Rec& r = recs[k];
						if (r.font == font && r.text_color == color_id)
							{
							r.font = 0;					// Mark record "used".
							r.text_color = -1;
							r.draw_text( lw );
							}
						}
					}
				}
			}
		}
	}

//-----------------------------------------------------------------------------
// draw
//-----------------------------------------------------------------------------
void MiscDrawList::draw()
	{
	if (num_recs > 0)
		{
		draw_icons();
		draw_text();
		empty();
		}
	}


//-----------------------------------------------------------------------------
// empty
//-----------------------------------------------------------------------------
void MiscDrawList::empty()
	{
	for (int i = 0; i < num_recs; i++)
		if (recs[i].text != 0)
			free( recs[i].text );
	num_recs = 0;
	}


//-----------------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------------
MiscDrawList::~MiscDrawList()
	{
	empty();
	free( recs );
	}


//-----------------------------------------------------------------------------
// Constructor
//-----------------------------------------------------------------------------
MiscDrawList::MiscDrawList()
	{
	num_recs = 0;
	max_recs = 16;
	recs = (Rec*) malloc( max_recs * sizeof( *recs ) );
	}

