/*
  MiscDragScrollWindow.m
  $Id$
  Copyright (c) 1997 Dwight D. Everhart.  All rights reserved.

  This notice may not be removed from this source code.

  This object is included in the MiscKit by permission from the author
  and its use is governed by the MiscKit license, found in the file
  "LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
  for a list of all applicable permissions and restrictions.
*/

#import <misckit/MiscDragScrollWindow.h>
#import <appkit/appkit.h>


#define arraySize(array) (sizeof (array) / sizeof ((array)[0]))



@implementation MiscDragScrollWindow



/*
  Note about getting the pasteboard types:  We mustn't get them in the
  enableScrolling... method, because that caused a nasty bug.  It turned
  out that asking the drag pasteboard for the types (with the "types"
  method) sometimes caused serious problems with the Workspace manager,
  including:  rejecting the drag, even though the app would've accepted it;
  aborting the drag and leaving the dragged image on the screen; and even
  crashing the Workspace and forcing a very ungraceful logout.

  Now the class registers all the standard drag types when it initializes.
  This should be sufficient for most apps.  Apps that use a custom
  pasteboard type and no standard types will have to use the setTypes:count:
  method before invoking enableScrolling:forDragSession.
*/

- initContent: (const NXRect *) contentRect style: (int) aStyle
  backing: (int) backingType buttonMask: (int) mask defer: (BOOL) flag
{
   NXAtom dragTypes[] =
   {
      NXAsciiPboardType,
      NXPostScriptPboardType,
      NXTIFFPboardType,
      NXRTFPboardType,
      NXSoundPboardType,
      NXFilenamePboardType,
      NXTabularTextPboardType,
      NXFontPboardType,
      NXRulerPboardType,
      NXFileContentsPboardType,
      NXColorPboardType,
      NXSelectionPboardType,
      NXDataLinkPboardType,
   };

   if (![super initContent: contentRect style: aStyle
	 backing: backingType buttonMask: mask defer: flag]) {
      return nil;
   }

   [self registerForDraggedTypes: dragTypes count: arraySize (dragTypes)];
   return self;
}



- free
{
   [self unregisterDraggedTypes];
   return [super free];
}



- setDragTypes: (const NXAtom *) types count: (int) typeCount
{
   [self registerForDraggedTypes: types count: typeCount];
   return self;
}



- enableScrolling: (View *) aView
  forDragSession: (id <NXDraggingInfo>) session
{
   if ((scrollingView != aView) ||
       (sequence != [session draggingSequenceNumber])) {
      scrollingView = aView;
      sequence = [session draggingSequenceNumber];
   }

   return self;
}



- disableScrolling
{
   scrollingView = nil;
   return self;
}



- (NXDragOperation) draggingEntered: (id <NXDraggingInfo>) sender
{
   if (scrollingView && ([sender draggingSequenceNumber] == sequence)) {
      return [self draggingUpdated: sender];
   } else {
      [self disableScrolling];
      return NX_DragOperationNone;
   }
}



- (NXDragOperation) draggingUpdated: (id <NXDraggingInfo>) sender
{
   NXEvent theEvent;

   /*
     We need to check the sequence number, because, if we don't, we may
     scroll a view that the cursor hasn't yet entered during this drag
     session.  This could cause, for instance, the view to suddenly jump
     to the top or bottom, because we would begin scrolling it as soon
     as the cursor enters the window, either from the top or the bottom.
   */
   if (scrollingView && ([sender draggingSequenceNumber] == sequence)) {
      theEvent = *[NXApp currentEvent];
      theEvent.location = [sender draggingLocation];
      [scrollingView autoscroll: &theEvent];
   }

   return NX_DragOperationNone;
}



- draggingExited: (id <NXDraggingInfo>) sender
{
   [self disableScrolling];
   return self;
}



@end

