//=============================================================================
//
//		Copyright (C) 1995-1996 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//		
//=============================================================================
//-----------------------------------------------------------------------------
// MiscBorderCell.M
//
//		Cell class used by MiscBorderView to manage column headings and row 
//		labels for MiscTableScroll.	 
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscBorderCell.M,v 1.3 96/05/05 10:55:10 sunshine Exp $
// $Log:		MiscBorderCell.M,v $
//	Revision 1.3  96/05/05	10:55:10  sunshine
//	Rewrote -drawSelf:inView: so that it makes only 1 trip to Window Server
//	instead of 8, and is now table driven.
//	
//	Revision 1.2  96/04/25	22:08:56  sunshine
//	Added missing -calcCellSize:inRect:.
//	
//	Revision 1.1  95/09/27	12:21:21  zarnuk
//	Initial revision
//-----------------------------------------------------------------------------
#import "MiscBorderCell.h"

extern "Objective-C" {
#import <appkit/Font.h>
#import <appkit/Text.h>
#import <dpsclient/wraps.h>
}

@implementation MiscBorderCell

//-----------------------------------------------------------------------------
// -initTextCell:
//-----------------------------------------------------------------------------
- initTextCell: (char const*) s
	{
	[super initTextCell:s];
	[self setBordered:NO];
	[self setWrap:NO];
	[self setAlignment: NX_CENTERED];
	[self setFont: [Font boldSystemFontOfSize:12.0 matrix:NX_FLIPPEDMATRIX]];
	return self;
	}


//-----------------------------------------------------------------------------
// -backgroundGray
//-----------------------------------------------------------------------------
- (float) backgroundGray
	{
	return cFlags1.state ? (1.0/6.0) : NX_DKGRAY;
	}


//-----------------------------------------------------------------------------
// -setTextAttributes:
//-----------------------------------------------------------------------------
- setTextAttributes:textObject
	{
	[super setTextAttributes:textObject];
	[textObject setTextGray:NX_WHITE];
	[textObject setBackgroundGray: [self backgroundGray]];
	return textObject;
	}


//-----------------------------------------------------------------------------
// -calcCellSize:inRect:
//-----------------------------------------------------------------------------
- calcCellSize:(NXSize*)theSize inRect:(NXRect const*)aRect
	{
	[super calcCellSize:theSize inRect:aRect];
	theSize->width += 3;
	theSize->height += 3;
	return self;
	}


//-----------------------------------------------------------------------------
// -drawInside:inView:
//-----------------------------------------------------------------------------
- drawInside:(NXRect const*)cellFrame inView:controlView
	{
	NXRect r;
	PSsetgray( [self backgroundGray] );
	NXSetRect( &r, NX_X(cellFrame)+2, NX_Y(cellFrame)+2,
				   NX_WIDTH(cellFrame) - 3, NX_HEIGHT(cellFrame) - 3 );
	NXRectFill( &r );
	[super drawInside:&r inView:controlView];
	return self;
	}


//-----------------------------------------------------------------------------
// -drawSelf:inView:
//
//		Draw light gray lines at top and left edges.  Draw white lines just
//		inside light gray lines at top and left edges.	Draw black lines at
//		bottom and right edges.	 Draw dark gray pixel at junction of white
//		and black lines (in lower-left and upper-right corners).
//
//		The lines are derived from a table of scalers and constants which are
//		applied to the cellFrame(cf) using the following formulas:
//				x = cf.x * xScaler + cf.maxx * maxxScaler + xConst
//				y = cf.y * yScaler + cf.maxy * maxyScaler + yConst
//				w = cf.w * wScaler + wConst
//				h = cf.h * hScaler + hConst
//
//-----------------------------------------------------------------------------
- drawSelf:(NXRect const*)cellFrame inView:aView
	{
	int const NUM_RECTS = 8;
	float const grays[ NUM_RECTS ] =
						{ NX_LTGRAY, NX_LTGRAY, NX_WHITE, NX_WHITE,
						  NX_BLACK, NX_BLACK, NX_DKGRAY, NX_DKGRAY };
	struct BC_Lines { char xs, maxxs, xc, ys, maxys, yc, ws, wc, hs, hc; };
	static BC_Lines const LINES[ NUM_RECTS ] =
		{
			{ 1, 0,	 0, 1, 0,  0, 0,  1, 1,	 0 }, // ltgray, left edge
			{ 1, 0,	 1, 1, 0,  0, 1, -1, 0,	 1 }, // ltgray, top edge
			{ 1, 0,	 1, 1, 0,  1, 0,  1, 1, -2 }, // white,	 left edge + 1
			{ 1, 0,	 2, 1, 0,  1, 1, -3, 0,	 1 }, // white,	 top edge + 1
			{ 0, 1, -1, 1, 0,  2, 0,  1, 1, -2 }, // black,	 right edge
			{ 1, 0,	 2, 0, 1, -1, 1, -3, 0,	 1 }, // black,	 bottom edge
			{ 1, 0,	 1, 0, 1, -1, 0,  1, 0,	 1 }, // dkgray, lower left corner
			{ 0, 1, -1, 1, 0,  1, 0,  1, 0,	 1 }  // dkgray, upper right corner
		};

	NXRect rects[ NUM_RECTS ];
	for (int i = 0; i < NUM_RECTS; i++)
		{
		BC_Lines const& r = LINES[i];
		NXSetRect( &(rects[i]),
				r.xs * NX_X(cellFrame) + r.maxxs * NX_MAXX(cellFrame) + r.xc,
				r.ys * NX_Y(cellFrame) + r.maxys * NX_MAXY(cellFrame) + r.yc,
				r.ws * NX_WIDTH(cellFrame) + r.wc,
				r.hs * NX_HEIGHT(cellFrame) + r.hc );
		}

	NXRectFillListWithGrays( rects, grays, NUM_RECTS );
	return [self drawInside:cellFrame inView:aView];
	}

@end
