/*
 *  $Id: MiscThreadedApp.h,v 1.1 1997/04/07 05:43:06 croehrig Exp $
 *
 *  MiscThreadedApp:   adds support for a multi-threaded application.
 *
 *  (c) 1997 Chris Roehrig <croehrig@House.ORG>
 *  All Rights Reserved
 *
 *  This notice may not be removed from this source code.
 *
 *  This object is included in the MiscKit by permission from the author
 *  and its use is governed by the MiscKit license, found in the file
 *  "License.rtf" in the MiscKit distribution.  Please refer to that file
 *  for a list of all applicable permissions and restrictions.
 *
 *  MiscThreadedApp is a subclass of Application that supports
 *  multithreaded operation in a simple way.   It contains the
 *  functionality of the OpenStep NSThread object, but also provides
 *  increased support for interaction with the AppKit.    The AppKit
 *  is not thread-safe and only the main thread can use it.  All other
 *  threads must message the main thread to perform any functions
 *  (drawing, etc) that involve the AppKit.   MiscThreadedApp does this
 *  by introducing the notion of callback methods.   This is a flexible
 *  mechanism whereby a thread can request that the main application
 *  thread invoke a method on its behalf.     This allows a thread to
 *  have virtually full access to the AppKit in a way that is easy to
 *  understand and use.
 *  
 *  The callback mechanism uses Mach messages to make requests to the
 *  main AppKit thread.   The main thread only receives the messages
 *  when it is in an event loop, so it is crucial that the main thread
 *  remains responsive to events. (This is only good user-interface
 *  design anyways.)
 *  
 *  MiscThreadedApp provides support for locking shared data or code
 *  via the MiscLock and MiscConditionLock classes.   These classes
 *  allow the main AppKit thread to "block" while waiting to acquire
 *  a lock, but still remain responsive to events so that other threads
 *  may still perform callbacks.
 *  
 *  There are three priorities for callback methods:  fast, safe, and
 *  slow.   Fast callbacks are used when a thread needs to display a
 *  result, or to change some shared data.    An example usage is
 *  
 *  while( running ){
 *      //... code to update a value 
 *
 *      // update the display to reflect the new value 
 *      [NXApp callbackTarget:textField 
 *              perform:@selector(setIntValue:) with:(id)value]; 
 *  }
 *  
 *  In this case, the thread requests the application (NXApp) to invoke
 *  textField's setIntValue: method with the parameter value.  Note
 *  that value needs to be cast to an id, much like using Object's
 *  perform:with: method.  Fast callbacks are invoked directly from
 *  the Mach message handler which is dispatched from within the
 *  DPSGetEvent function.   Because of this, fast callback methods
 *  cannot get events (since DPSGetEvent is not re-entrant).   In
 *  particular, this means that fast callback methods should not use
 *  MiscLocks since acquiring a lock from within the main thread involves
 *  running a modal-event loop (see MiscLock).    Fast callback messages
 *  are sent with a priority of NX_MODALRESPTHRESHOLD + 1 and will
 *  therefore get through any user modal event loops.
 *  
 *  Safe callback methods are able to use the full functionality of
 *  the AppKit and run their own modal event loops if desired.   Safe
 *  callback methods are not run directly from the Mach message handler.
 *  Instead, the Mach message handler posts an EV_CALLBACK event to
 *  the application's event queue.   This is an NX_APPDEFINED event,
 *  handled by MiscThreadedApp's applicationDefined: method which
 *  performs the actual call-back.     Because the safe callback method
 *  is dispatched in response to a regular application event, it may
 *  do anything you might normally do in response to an event.    Safe
 *  callback methods also are able to use MiscLocks and MiscConditionLocks
 *  to access shared data.   Safe callback messages are sent with a
 *  priority of NX_MODALRESPTHRESHOLD + 1 and will therefore get through
 *  any user modal event loops.
 *  
 *  Slow callbacks are like safe callbacks, but are sent with a message
 *  priority of NX_BASETHRESHOLD and will therefore not interrupt any
 *  high priority things a user might be doing (attention panels,
 *  scrolling, other modal loops).
 *  
 *   Fast, safe and slow callback methods can be blocking or non-blocking.
 *   The callbackTarget:perform: methods are non-blocking: they don't
 *   wait for the invoked method to return.  The callbackAndWaitTarget:perform:
 *   methods block until the main thread has executed the method and
 *   returned the result.   The result is an id, but can be typecast
 *   to any int-sized type.
 *  
 *  If a callback method is invoked from the main thread of execution,
 *  the Object's perform: method is directly used to invoke the method
 *  and no messages are sent.
 *
 *
 */     
#import <appkit/Application.h>
#import <mach/cthreads.h>


@interface MiscThreadedApp : Application
{
}


- (void)detachNewThreadSelector:	(SEL)aSelector 
			toTarget:	(id)aTarget 
			withObject:	(id)anArgument;
/*
 *  Start the method in a new thread of execution.  The method is called
 *  as in [aTarget perform:aSelector with:anArgument].   Any thread can
 *  invoke this method.
 */


- (void)exitCurrentThread;
/*
 *  Exits the caller's thread of execution.
 */


- (void)sleepCurrentThread:(int)msec;
/*
 *  Performs a context switch and puts the current thread to sleep for
 *  msec milliseconds.   Use this as an alternative to usleep
 *  (which is not thread-safe).
 */


/*
 * The following callback methods cause a method to be executed under the
 * App's main thread of execution.  They behave similarly to Object's 
 * perform:with: method except a target is also specified.   
 * Any int-sized argument can be used and typecast to an id.
 *
 *  If a callback method is invoked from the main thread of execution,
 *  the perform: method is directly used and no messages are sent.
 *
 * Typical usage is:
 *     [NXApp callback:self perform:@selector(update:) with:self];
 */

- (id)callbackAndWaitTarget:target perform:(SEL)aSel;
- (id)callbackAndWaitTarget:target perform:(SEL)aSel with:anObject;
- (void)callbackTarget:target perform:(SEL)aSel;
- (void)callbackTarget:target perform:(SEL)aSel with:anObject;

- (id)safeCallbackAndWaitTarget:target perform:(SEL)aSel;
- (id)safeCallbackAndWaitTarget:(id)target perform:(SEL)aSel with:(id)anObject;
- (void)safeCallbackTarget:target perform:(SEL)aSel;
- (void)safeCallbackTarget:(id)target perform:(SEL)aSel with:(id)anObject;

- (id)slowCallbackAndWaitTarget:target perform:(SEL)aSel;
- (id)slowCallbackAndWaitTarget:(id)target perform:(SEL)aSel with:(id)anObject;
- (void)slowCallbackTarget:target perform:(SEL)aSel;
- (void)slowCallbackTarget:(id)target perform:(SEL)aSel with:(id)anObject;


@end


// Private methods in a category for MiscLock...
@interface MiscThreadedApp(MiscLock) 
- (void)lock: (mutex_t)mutex;
- (void)lock: (mutex_t)mutex whenCondition: (condition_t)cond
		withData:(volatile int *)data is:(int)value;
- (void)threadDidUnlock:sender;
@end
